#pragma once 

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <syslog.h>
#include <stdbool.h>
#include <sys/types.h>
#ifdef _x86_64
#include <asm-x86_64/types.h>
#endif

#include <sys/socket.h>
#include <sys/un.h>

#include <arpa/inet.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>
#include <netinet/ip.h>
#include <netinet/ip6.h>

#include <glib.h>
#include <pthread.h>

#include <libgen.h>
#include <linux/netfilter.h>
#include <libnetfilter_queue/libnetfilter_queue.h>
#include <libnetfilter_queue/linux_nfnetlink_queue.h>

#define ATOMIC_LOAD(var)    __atomic_load_n((var), __ATOMIC_RELAXED)
#define ATOMIC_STORE(var, value)   __atomic_store_n((var), (value), __ATOMIC_RELAXED)

#define MDE_NETFILTER_SOCKET_PATH "/var/run/mde_netfilter.sock"


#define LOG_MSG(level, str, fmt, ...)  \
        syslog(level, "%s: %s: " fmt, str, __FUNCTION__, ##__VA_ARGS__)

#define INFO(fmt, ...)   LOG_MSG(LOG_INFO,    "INFO",   fmt, ##__VA_ARGS__)
#define WARN(fmt, ...)   LOG_MSG(LOG_WARNING, "WARN",   fmt, ##__VA_ARGS__)
#define ERROR(fmt, ...)  LOG_MSG(LOG_ERR,     "ERROR",  fmt, ##__VA_ARGS__)
#define DEBUG(fmt, ...)  LOG_MSG(LOG_DEBUG,   "DEBUG",  fmt, ##__VA_ARGS__)
#define NOTICE(fmt, ...) LOG_MSG(LOG_NOTICE,  "NOTICE", fmt, ##__VA_ARGS__)

#define CHECK_ERROR(cond, fmt, ...)     \
    do {                                \
        if (cond)                       \
        {                               \
            ERROR(fmt, ##__VA_ARGS__);  \
            goto exit;                  \
        }                               \
} while (0)

/* So far it can support upto two clients */
#define MDE_MAX_CLIENTS 2

/* Network event Type */
enum conn_event_type {
    OUTBOUND_PRECONNECT = 1,             /* outgoing connection initiation*/
    INBOUND_PRECONNECT,                  /* incoming connection inititation */
    MAX_EVENT,
};

/* Network connection identification related data */
struct conn_identity_data {
    struct sockaddr_storage src;         /* source ip */
    struct sockaddr_storage dst;         /* destination ip */
    enum conn_event_type event_type;     /* network connection type */
    uint32_t event_id;                   /* event id */
    uint8_t protocol;                    /* L3 protocol */
};

/* verdict information from client */
typedef struct _verdict {
    uint32_t packetid;                   /* packet tracking id for client */
    int verdict;                         /* verdict received from client */
} packet_verdict;

/* client context */
typedef struct _client_context {
    int version;                         /* client version */
    int fd;                              /* socket fd of the client */
    uint32_t queue_num;                  /* iptables queue_num */
    uint32_t mask;                       /* iptables mask */
} client_ctx;

/*
 * A wrapper structure to keep netfilter queue handle and FD for getting
 * packets from queue library/kernel module.
 */
struct _net_queue {
   struct nfq_handle   *handle;         /* Netfilter queue connection handle */
   struct nfq_q_handle *qhandle;        /* Netfilter queue handle */
   int qfd;                             /* Netfilter queue fd */
};

/*
 * Global data structure that contains netfilter related structures, listen
 * start/stop hash cache and thread pool structure  for delivering events.
 */
struct net_session {
    struct _net_queue  queue_ctx;         /* NFQ handle */
    pthread_mutex_t queue_lock;           /* lock to sync nfq send/recv */
    client_ctx clients[MDE_MAX_CLIENTS];  /* connected clients information */
    int timerfd;                          /* timerfd - on timeout auto approve */
    pthread_mutex_t pkt_hashtable_lock;   /* lock to protect hashtable */
    GHashTable *pkt_hashtable;            /* data strucutre to hold outstanding packets */
};

/* object in the packet hashtable data structue */
struct outstanding_pkt {
    uint32_t packetid;                   /* packet tracking id for client */
    uint64_t pkt_time;                   /* packet timestamp */
    uint32_t pkt_mark;                   /* packet mark require while responding the verdict */
};

extern int netfilter_platform_init(struct net_session *sess);
extern void* netfilter_event_handler(void *arg);
extern void* client_verdict_receiver(void *arg);
extern void net_cleanup(struct net_session *sess);
extern uint64_t get_time_in_millis();
extern int create_named_socket(const char*, size_t);